#include<stdio.h>
#include<stdlib.h>
#include<assert.h>
#include<stddef.h>
#include<malloc.h>
#include<memory.h>

//! \file memory4m.cpp
//      memory4.cpp
//   ,     
// (c) Proteus       lawnmower-man@mail.ru

#define MEM_ALIGN	4			//!<    
#define PAGE_SIZE   1024		//!<    

#define BANK_SIG 0x12345678	//!<   
#define ITEM_SIG 0xABCD		//!<   
#define LIST_LINK 0			//!<   
#define FREE_LINK 1			//!<   

typedef unsigned int   uint;
typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned int  dword;
typedef word index_t;			//!<  

//! 
struct stat_st{
	int pages_count;		//!<  
	int free_pages_count;	//!<     
	int all_space;			//!<    
	int free_space;			//!<    
	int lost_space;			//!<    
	int obj_count;			//!<   
};


#pragma pack(push,1)

//!  
class StubAllocator{
	uint item_size;		//!<   
	//!  
	union item_st{
		struct {
			index_t next_free;	//!<   .  
			word    sig;		//!<   (  )
		};
		byte data[1];			//!<     
	};
	//!  
	struct bank_st{
		bank_st *prev[2];		//!<    (  )
		bank_st *next[2];		//!<    (  )
		index_t  free_index;	//!<      
		index_t  free_count;	//!<   
		dword sig;				//!<  
		byte data[1];			//!<  
	};
	bank_st *bank_list;										//!<   
	bank_st *free_list;										//!<   
	void NewBank();											//!<    
	bank_st *ListInsert(bank_st*,bank_st*,unsigned int i);	//!<      
	bank_st *ListRemove(bank_st*,unsigned int i);			//!<     
	//!    
	const uint PageCount()
	{
		return (PAGE_SIZE-sizeof(bank_st))/(item_size+1);
	}
	//!     
	//! \param index    
	int ElemOffset(uint index)
	{
		return (item_size+1)*index;
	}
	//!    
	//! \param bank   
	//! \param index    
	item_st &Elem(bank_st *bank,uint index)
	{
		return *(item_st*)&bank->data[ElemOffset(index)];
	}
	//!    
	//! \param bank   
	//! \param index    
	byte &ElemID(bank_st *bank,uint index)
	{
		return bank->data[ElemOffset(index)+item_size];
	}
public:
	StubAllocator(uint);
	~StubAllocator();
	void *New();			//!<  
	bool Free(void*);		//!<  
	void Clear();			//!<     
	bool isCorrect();		//!<    (   )
	void Stat(stat_st*);	//!<   
};

//!     
class Allocator{
	StubAllocator **array;	//!<   
	uint capablity;			//!<   
	void Resize(uint);		//!<    
public:
	Allocator();
	~Allocator();
	void *New(uint size);			//!<   
	void Free(void *p,uint size);	//!<  
	void Clear();					//!<  
	void Stat();
};

#pragma pack(pop)


//! \param size    
StubAllocator::StubAllocator(uint size):item_size(size)
{
	assert(isCorrect());
	bank_list=0;
	free_list=0;
}


StubAllocator::~StubAllocator()
{
	Clear();
}

//!  .       
void StubAllocator::Clear()
{
	bank_st *next;
	free_list=0;
	if(bank_list) bank_list->prev[LIST_LINK]->next[LIST_LINK]=0;
	while(bank_list) {
		next=bank_list->next[LIST_LINK];
		delete bank_list;
		bank_list=next;
	}
}

//!     
//! \param list 
//! \param bank  
//! \param i        (   )
//! \result    (      )
StubAllocator::bank_st *StubAllocator::ListInsert(bank_st *list,bank_st *bank,unsigned int i)
{
	if(!list) list=bank->next[i]=bank->prev[i]=bank;
	else {
		bank->prev[i]=list->prev[i];
		bank->next[i]=list;
		list->prev[i]=bank;
		bank->prev[i]->next[i]=bank;
	}
	return list;
}

//!     
//! \param list  (  )
//! \param i        (   )
//! \result   
StubAllocator::bank_st *StubAllocator::ListRemove(bank_st *list,unsigned i)
{
	bank_st *prev;
	if(!list || list->next[i]==list) return 0;
	prev=list->prev[i];
	list=list->next[i];
	prev->next[i]=list;
	list->prev[i]=prev;
	return list;
}

//!      
void StubAllocator::NewBank()
{
	//  
	bank_st *bank=(bank_st*)malloc(PAGE_SIZE);
	//     (     )
	bank_list=ListInsert(bank_list,bank,LIST_LINK);
	free_list=ListInsert(free_list,bank,FREE_LINK);
	free_list->sig=BANK_SIG;
	free_list->free_index=0;
	free_list->free_count=(index_t)PageCount();
	//    
	index_t n;
	for(n=0; n<free_list->free_count-1; n++) 
		Elem(free_list,n).next_free=n+1;
		Elem(free_list,n).next_free=(index_t)-1;
	for(n=0; n<free_list->free_count; n++) {
		ElemID(free_list,n)=(byte)(n>(byte)-1?-1:n);
		if(item_size>=sizeof(item_st)) Elem(free_list,n).sig=ITEM_SIG;
	}
}

//!   (     )
//! result true  
bool StubAllocator::Free(void *obj)
{
	if(!obj) return false;
	//   
	unsigned int index=0;
	byte *ref=(byte*)obj+item_size;
	while(*ref) {
		index+=*ref;
		ref-=*ref*(item_size+1);
		if(index>PAGE_SIZE-sizeof(bank_st)) return true;
	}
	bank_st *bank=(bank_st*)(ref-offsetof(bank_st,data)-item_size);
	if(bank->sig!=BANK_SIG) return true;
	//     ,      
	if(bank->free_index==(index_t)-1) free_list=ListInsert(free_list,bank,FREE_LINK);
	//      
	item_st *item=(item_st*)obj;
	item->next_free=bank->free_index;
	if(item_size>=sizeof(item_st)) item->sig=ITEM_SIG;
	bank->free_index=index;
	bank->free_count++;
	//      ,   )
	if(bank->free_count==PageCount()) {
		bank_list=ListRemove(bank,LIST_LINK);
		bank_st *temp=ListRemove(bank,FREE_LINK);
		if(bank==free_list) free_list=temp; //     if(!temp) free_list=0,     
		delete bank;
	}
	//            (     )
	else if(bank->free_count<free_list->free_count) free_list=bank;
	return false;
}


//!   
//! \param 0  
void* StubAllocator::New()
{
	//    ,  
	if(!free_list) NewBank();
	//     
	item_st *item=&Elem(free_list,free_list->free_index);
	if(item_size>=sizeof(item_st) && item->sig!=ITEM_SIG) return 0;
	free_list->free_index=item->next_free;
	free_list->free_count--;
	//      ,      
	if(free_list->free_index==(index_t)-1) free_list=ListRemove(free_list,FREE_LINK);
	// ..
	return item->data;
}

//!    (     )
//! \result false    
bool StubAllocator::isCorrect()
{
	return (index_t)-1>PageCount() && item_size<PAGE_SIZE/3;
}


void StubAllocator::Stat(stat_st *stat)
{
#ifdef _DEBUG
	int t;
	bank_st *bank=bank_list;
	if(bank) do {
		t=PageCount()*(item_size+1);
		stat->pages_count++;
		stat->all_space+=t;
		stat->lost_space+=PAGE_SIZE-sizeof(bank_st)-t;
		bank=bank->next[LIST_LINK];
	}while(bank!=bank_list);
	bank=free_list;
	if(bank) do {
		stat->free_pages_count++;
		stat->free_space+=bank->free_count*(item_size+1);
		bank=bank->next[FREE_LINK];
	}while(bank!=free_list);
#endif
}


Allocator::Allocator()
{
	array=0; capablity=0;
}


Allocator::~Allocator()
{
	for(uint n=0; n<capablity; n++) delete array[n];
	free(array);
}


//! \param size    (  ,   )
void Allocator::Resize(uint size)
{
	if(size<capablity) return;
	uint up=capablity;
	while(up<=size) up+=1+up/4;
	array=(StubAllocator**)realloc(array,up*sizeof(StubAllocator*));
	while(capablity<up) array[capablity++]=0;
}


//! \param size   
//! \result   
void *Allocator::New(uint size)
{
	if(size>PAGE_SIZE/5) return malloc(size);
	int n=(size+MEM_ALIGN-1+sizeof(index_t))/MEM_ALIGN;
	size=n*MEM_ALIGN-sizeof(index_t);
	Resize(n);
	if(!array[n]) array[n]=new StubAllocator(size);
	return array[n]->New();
}


//! \param p  
//! \param size  
void Allocator::Free(void *p,uint size)
{
	if(size>PAGE_SIZE/5) free(p); else {
		int n=(size+MEM_ALIGN-1+sizeof(index_t))/MEM_ALIGN;
		if(array[n]->Free(p)) delete array[n],array[n]=0;
	}
}


void Allocator::Stat()
{
#ifdef _DEBUG
	uint n;
	stat_st stat; memset(&stat,0,sizeof(stat));
	for(n=0; n<capablity; n++) if(array[n]) array[n]->Stat(&stat),stat.obj_count++;
	printf("[stat]\n");
	printf("pages count     : %d\n",stat.pages_count);		//  
	printf("free pages count: %d\n",stat.free_pages_count);	//     
	printf("all space       : %d\n",stat.all_space);		//    
	printf("free space      : %d\n",stat.free_space);		//    
	printf("lost space      : %d\n",stat.lost_space);		//    
	printf("obj count       : %d\n",stat.obj_count);		//   
#endif
}



#define COUNT 15000		//    
#define SIZE  300		//    


int main()
{
// 
	int n,i;
	struct{int size; int *ref;}test[COUNT]; 
//------------------------------------------------
	memset(test,0,sizeof(test));
	StubAllocator obj1(sizeof(int));
	for(n=0; n<COUNT*1000; n++) {
		i=rand()%COUNT;
		printf("[1]pass: %d(%02d%%)\r",n,(n+1)*100/(COUNT*1000));
		if(test[i].ref) {
			if(*test[i].ref==i) throw 1;
			obj1.Free(test[i].ref),test[i].ref=0;
		}else{
			test[i].ref=(int*)obj1.New();
			if(!test[i].ref) throw 2;
		}
	}
	putchar('\n');
//------------------------------------------------
	Allocator obj2;
	memset(test,0,sizeof(test));;
	for(n=0; n<COUNT*1000; n++) {
		i=rand()%COUNT;
		printf("[2]pass: %d(%02d%%)\r",n,(n+1)*100/(COUNT*1000));
		if(test[i].ref) {
			if(*test[i].ref==i) throw 3;
			obj2.Free(test[i].ref,test[i].size),test[i].ref=0;
		}else{
			test[i].size=rand()%SIZE+1;
			test[i].ref=(int*)obj2.New(test[i].size);
			if(!test[i].ref) throw 4;
		}
	}
	putchar('\n');
	for(n=i=0; i<COUNT; i++) if(test[i].ref) n+=test[i].size;
	obj2.Stat();
	printf("use space: %d\n",n);
	return 0;
}

